%Variational learning for the basic example fo Lecture 7. 
%by Umut Orguner on March 02, 2011.
clear
close all

myinf=1e10;% a very large number
linewidth=2;
fontsize=15;

sigmav=sqrt(.1);%process noise std
sigmae=sqrt(.1);%measurement noise std
sigma0=sqrt(.1);%initial state std
sigmatheta=100;%prior std

thetatrue=0.9;%True parameter value

%state model
xmodel.A=thetatrue;
xmodel.B=1;
xmodel.Q=sigmav^2;
xmodel.x0=0;
xmodel.P0=sigma0^2;

%measurement model
ymodel.C=0.5;
ymodel.D=1;
ymodel.R=sigmae^2;


Ndata=5;%Number of measurement data
Niter=20;%number of iterations

[y,x]=generatedata(xmodel,ymodel,Ndata);%generate true data



THETA=-2:0.01:2;%theta values to evaluate the true posterior
Ntheta=length(THETA);


xmodeltemp=xmodel;
logposterior=zeros(1,Ntheta);
for i=1:Ntheta,
    i
    theta=THETA(i);
    xmodeltemp.A=theta;
    [~,~,~,~,logy]=kf(xmodeltemp,ymodel,y);%calculate p(y_0:N|theta)=sum_k log p(y_k|y_0:k-1,theta)
    logposterior(i)=logy+log(normpdf(theta,0,sigmatheta));%calculate the true log posterior (up to an additive constant) by adding the prior 
end
posterior=exp(logposterior);%get the true posterior up to a multiplicative constant
normposterior=(posterior/sum((THETA(2)-THETA(1))*posterior));%normalize to get the true posterior

%plot the true posterior
figure(1)
hposterior=plot(THETA,normposterior,'b','linewidth',linewidth);
hold on
title(['Number of data points: ',num2str(Ndata),' iter no: ', num2str(0) ],'FontSize',fontsize);


maxposterior=max(normposterior);
minposterior=min(normposterior);
%plot true theta
htrue=plot([thetatrue,thetatrue],[minposterior maxposterior],'k','linewidth',linewidth);


[xmodelEM, ymodelEM]=doublemodel(xmodel,ymodel);%get the augmented model for EM 
[xmodelVB,ymodelVB]=doublemodel(xmodel,ymodel);%get the augmented model for VB




thetahatEM=zeros(1,Niter+1);%EM estimates
tempthetaEM=0;%Current EM estimate

thetahatVB=zeros(1,Niter+1);%VB estimates
meanthetaVB=0;%Current VB estimate
varthetaVB=sigmatheta^2;

%Definition of extra measurements for the VB
yextra=mat2cell(zeros(Ndata+1,1),ones(1,Ndata+1),1);
yextra{end}=[];
CextraVB=mat2cell(repmat([1,0],Ndata+1,1),ones(1,Ndata+1),2);
CextraVB{end}=[];
DextraVB=mat2cell(ones(Ndata+1,1),ones(1,Ndata+1),1);
DextraVB{end}=[];

ymodelextra.C=CextraVB;
ymodelextra.D=DextraVB;



%plot initial estimates
hEM=plot([tempthetaEM,tempthetaEM],[minposterior maxposterior],'r','linewidth',linewidth);
hVB=plot([meanthetaVB,meanthetaVB],[minposterior maxposterior],'g','linewidth',linewidth);
hVBdensity=plot(THETA,(normpdf(THETA,meanthetaVB,sqrt(varthetaVB))),'g','linewidth',linewidth) ;
legend([hposterior,htrue,hEM,hVB],'True posterior','True \theta','EM','VB');
pause


for i=1:Niter,
    pause(1)    
    %%%%EM recursion
    xmodelEM.A=[tempthetaEM 0; 1,0];
    [xs,Ps ~]=ks(xmodelEM,ymodelEM,y);%Kalman smoother
    tempthetaEM=(sum(Ps(1,2,2:Ndata+1))+sum(xs(1,2:Ndata+1).*xs(2,2:Ndata+1)))/(sum(Ps(2,2,2:Ndata+1))+sum(xs(2,2:Ndata+1).^2));
    thetahatEM(i+1)=tempthetaEM;
    delete(hEM);
    hEM=plot([tempthetaEM,tempthetaEM],[minposterior maxposterior],'r','linewidth',linewidth);%plot current EM estimate
    title(['Number of data points: ',num2str(Ndata),' iter no: ', num2str(i) ],'FontSize',fontsize);
    %%%%End of EM recursion
    
    %%VB recursion
    xmodelVB.A=[meanthetaVB 0; 1,0];%set the current model
    ymodelextra.R=mat2cell(sigmav^2*ones(Ndata+1,1)/varthetaVB,ones(1,Ndata+1),1);
    ymodelextra.R{end}=[];
    [xs,Ps ~]=ks(xmodelVB,ymodelVB,y,ymodelextra,yextra);%Kalman smoother    
    varthetaVB=1/((sum(Ps(2,2,2:Ndata+1))+sum(xs(2,2:Ndata+1).^2))/(sigmav^2)+1/sigmatheta^2);%calculate the statistics
    meanthetaVB=varthetaVB*(sum(Ps(1,2,2:Ndata+1))+sum(xs(1,2:Ndata+1).*xs(2,2:Ndata+1)))/sigmav^2;
    thetahatVB(i+1)=meanthetaVB;
    delete(hVB);
    delete(hVBdensity);
    hVB=plot([meanthetaVB,meanthetaVB],[minposterior maxposterior],'g','linewidth',linewidth);%plot current VB theta estimate (mean of the approximate theta posterior)
    hVBdensity=plot(THETA,(normpdf(THETA,meanthetaVB,sqrt(varthetaVB))),'g','linewidth',linewidth) ;%plot current approximate theta posterior   
    %End of VB recursion
    
    legend([hposterior,htrue,hEM,hVB],'True posterior','True \theta','EM','VB')
end
hold off

figure(2)
htrue=plot(0:Niter,thetatrue*ones(1,Niter+1),'b','linewidth',linewidth);
hold on
hEM=plot(0:Niter,thetahatEM,'r','linewidth',linewidth)
hVB=plot(0:Niter,thetahatVB,'g','linewidth',linewidth)
xlabel('Iteration No')
ylim([-0.1 1])
legend([htrue,hEM,hVB],'true','EM','VB')
hold off